import { createSessionCookie } from "./cookie.js";
import { logError } from "../utils/log.js";
import { generateHashWithScrypt, generateRandomString, validateScryptHash } from "../utils/crypto.js";
import { LuciaError } from "./error.js";
import { parseCookie } from "../utils/cookie.js";
import { getSessionFromDatabaseData } from "./session.js";
export { SESSION_COOKIE_NAME } from "./cookie.js";
export const lucia = (configs) => {
    return new Auth(configs);
};
const validateConfigurations = (configs) => {
    const isAdapterProvided = configs.adapter;
    if (!isAdapterProvided) {
        logError('Adapter is not defined in configuration ("config.adapter")');
        process.exit(1);
    }
};
export class Auth {
    adapter;
    generateUserId;
    sessionCookie;
    sessionTimeout;
    env;
    hash;
    autoDatabaseCleanup;
    transformUserData;
    csrfProtection;
    constructor(configs) {
        validateConfigurations(configs);
        const defaultSessionCookieOption = {
            sameSite: "lax",
            path: "/"
        };
        if ("user" in configs.adapter) {
            if ("getSessionAndUserBySessionId" in configs.adapter.user) {
                delete configs.adapter.user.getSessionAndUserBySessionId;
            }
            if ("getSessionAndUserBySessionId" in configs.adapter.session) {
                delete configs.adapter.session.getSessionAndUserBySessionId;
            }
        }
        this.adapter =
            "user" in configs.adapter
                ? { ...configs.adapter.user, ...configs.adapter.session }
                : configs.adapter;
        this.generateUserId = configs.generateCustomUserId ?? (async () => null);
        this.env = configs.env;
        this.csrfProtection = configs.csrfProtection ?? true;
        this.sessionTimeout = {
            activePeriod: configs.sessionTimeout?.activePeriod ?? 1000 * 60 * 60 * 24,
            idlePeriod: configs.sessionTimeout?.idlePeriod ?? 1000 * 60 * 60 * 24 * 14
        };
        this.autoDatabaseCleanup = configs.autoDatabaseCleanup ?? true;
        this.transformUserData = ({ id, hashed_password, provider_id, ...attributes }) => {
            const transform = configs.transformUserData ??
                (({ id }) => {
                    return {
                        userId: id
                    };
                });
            return transform({ id, ...attributes });
        };
        this.sessionCookie = configs.sessionCookie ?? [defaultSessionCookieOption];
        this.hash = {
            generate: configs.hash?.generate ?? generateHashWithScrypt,
            validate: configs.hash?.validate ?? validateScryptHash
        };
    }
    getUser = async (userId) => {
        const databaseUser = await this.adapter.getUser(userId);
        if (!databaseUser)
            throw new LuciaError("AUTH_INVALID_USER_ID");
        const user = this.transformUserData(databaseUser);
        return user;
    };
    getUserByProviderId = async (provider, identifier) => {
        const providerId = `${provider}:${identifier}`;
        const databaseUser = await this.adapter.getUserByProviderId(providerId);
        if (!databaseUser)
            throw new LuciaError("AUTH_INVALID_PROVIDER_ID");
        const user = this.transformUserData(databaseUser);
        return user;
    };
    getSessionUser = async (sessionId) => {
        if (sessionId.length !== 40)
            throw new LuciaError("AUTH_INVALID_SESSION_ID");
        let userData;
        let sessionData;
        if (this.adapter.getSessionAndUserBySessionId !== undefined) {
            const databaseUserSession = await this.adapter.getSessionAndUserBySessionId(sessionId);
            if (!databaseUserSession)
                throw new LuciaError("AUTH_INVALID_SESSION_ID");
            userData = databaseUserSession.user;
            sessionData = databaseUserSession.session;
        }
        else {
            sessionData = await this.adapter.getSession(sessionId);
            userData = sessionData ? await this.adapter.getUser(sessionData.user_id) : null;
        }
        if (!sessionData)
            throw new LuciaError("AUTH_INVALID_SESSION_ID");
        const session = getSessionFromDatabaseData(sessionData);
        if (!session) {
            if (this.autoDatabaseCleanup) {
                await this.adapter.deleteSession(sessionId);
            }
            throw new LuciaError("AUTH_INVALID_SESSION_ID");
        }
        if (!userData)
            throw new LuciaError("AUTH_INVALID_USER_ID");
        return {
            user: this.transformUserData(userData),
            session
        };
    };
    createUser = async (provider, identifier, options) => {
        const providerId = `${provider}:${identifier}`;
        const attributes = options?.attributes ?? {};
        const userId = await this.generateUserId();
        const hashedPassword = options?.password ? await this.hash.generate(options.password) : null;
        const userData = await this.adapter.setUser(userId, {
            providerId,
            hashedPassword: hashedPassword,
            attributes
        });
        const user = this.transformUserData(userData);
        return user;
    };
    updateUserAttributes = async (userId, attributes) => {
        const [userData] = await Promise.all([
            this.adapter.updateUser(userId, {
                attributes
            }),
            this.autoDatabaseCleanup ? await this.deleteDeadUserSessions(userId) : null
        ]);
        const user = this.transformUserData(userData);
        return user;
    };
    updateUserProviderId = async (userId, provider, identifier) => {
        const providerId = `${provider}:${identifier}`;
        const [userData] = await Promise.all([
            this.adapter.updateUser(userId, {
                providerId
            }),
            this.autoDatabaseCleanup ? await this.deleteDeadUserSessions(userId) : null
        ]);
        const user = this.transformUserData(userData);
        return user;
    };
    updateUserPassword = async (userId, password) => {
        const hashedPassword = password ? await this.hash.generate(password) : null;
        const [userData] = await Promise.all([
            this.adapter.updateUser(userId, {
                hashedPassword
            }),
            this.invalidateAllUserSessions(userId)
        ]);
        const user = this.transformUserData(userData);
        return user;
    };
    deleteUser = async (userId) => {
        await this.adapter.deleteSessionsByUserId(userId);
        await this.adapter.deleteUser(userId);
    };
    authenticateUser = async (provider, identifier, password) => {
        const providerId = `${provider}:${identifier}`;
        const databaseData = await this.adapter.getUserByProviderId(providerId);
        if (!databaseData)
            throw new LuciaError("AUTH_INVALID_PROVIDER_ID");
        if (!databaseData.hashed_password)
            throw new LuciaError("AUTH_INVALID_PASSWORD");
        if (databaseData.hashed_password.startsWith("$2a"))
            throw new LuciaError("AUTH_OUTDATED_PASSWORD");
        const isValid = await this.hash.validate(password, databaseData.hashed_password);
        if (!isValid)
            throw new LuciaError("AUTH_INVALID_PASSWORD");
        const user = this.transformUserData(databaseData);
        return user;
    };
    getSession = async (sessionId) => {
        if (sessionId.length !== 40)
            throw new LuciaError("AUTH_INVALID_SESSION_ID");
        const databaseSession = await this.adapter.getSession(sessionId);
        if (!databaseSession)
            throw new LuciaError("AUTH_INVALID_SESSION_ID");
        const session = getSessionFromDatabaseData(databaseSession);
        if (!session) {
            if (this.autoDatabaseCleanup) {
                await this.adapter.deleteSession(sessionId);
            }
            throw new LuciaError("AUTH_INVALID_SESSION_ID");
        }
        return session;
    };
    validateSession = async (sessionId) => {
        const session = await this.getSession(sessionId);
        if (session.state === "active")
            return session;
        const renewedSession = await this.renewSession(sessionId);
        return renewedSession;
    };
    validateSessionUser = async (sessionId) => {
        const { session, user } = await this.getSessionUser(sessionId);
        if (session.state === "active")
            return { session, user };
        const renewedSession = await this.renewSession(sessionId);
        return {
            session: renewedSession,
            user
        };
    };
    generateSessionId = () => {
        const sessionId = generateRandomString(40);
        const activePeriodExpires = new Date(new Date().getTime() + this.sessionTimeout.activePeriod);
        const idlePeriodExpires = new Date(activePeriodExpires.getTime() + this.sessionTimeout.idlePeriod);
        return [sessionId, activePeriodExpires, idlePeriodExpires];
    };
    createSession = async (userId) => {
        const [sessionId, activePeriodExpires, idlePeriodExpires] = this.generateSessionId();
        await Promise.all([
            this.adapter.setSession(sessionId, {
                userId,
                expires: activePeriodExpires.getTime(),
                idlePeriodExpires: idlePeriodExpires.getTime()
            }),
            this.autoDatabaseCleanup ? await this.deleteDeadUserSessions(userId) : null
        ]);
        return {
            userId,
            activePeriodExpires,
            sessionId,
            idlePeriodExpires,
            state: "active",
            isFresh: true
        };
    };
    renewSession = async (sessionId) => {
        if (sessionId.length !== 40)
            throw new LuciaError("AUTH_INVALID_SESSION_ID");
        const sessionData = await this.adapter.getSession(sessionId);
        if (!sessionData)
            throw new LuciaError("AUTH_INVALID_SESSION_ID");
        const session = getSessionFromDatabaseData(sessionData);
        if (!session) {
            if (this.autoDatabaseCleanup) {
                await this.adapter.deleteSession(sessionId);
            }
            throw new LuciaError("AUTH_INVALID_SESSION_ID");
        }
        const [renewedSession] = await Promise.all([
            await this.createSession(session.userId),
            this.autoDatabaseCleanup ? await this.deleteDeadUserSessions(session.userId) : null
        ]);
        return renewedSession;
    };
    invalidateSession = async (sessionId) => {
        await this.adapter.deleteSession(sessionId);
    };
    invalidateAllUserSessions = async (userId) => {
        await this.adapter.deleteSessionsByUserId(userId);
    };
    deleteDeadUserSessions = async (userId) => {
        const sessions = await this.adapter.getSessionsByUserId(userId);
        const currentTime = new Date().getTime();
        const deadSessionIds = sessions
            .filter((val) => val.idle_expires < currentTime)
            .map((val) => val.id);
        if (deadSessionIds.length === 0)
            return;
        await this.adapter.deleteSession(...deadSessionIds);
    };
    validateRequestHeaders = (request) => {
        const cookies = parseCookie(request.headers.get("cookie") ?? "");
        const sessionId = cookies.auth_session ?? "";
        const checkForCsrf = request.method !== "GET" && request.method !== "HEAD";
        if (checkForCsrf && this.csrfProtection) {
            const origin = request.headers.get("Origin");
            if (!origin)
                throw new LuciaError("AUTH_INVALID_REQUEST");
            const url = new URL(request.url);
            if (url.origin !== origin)
                throw new LuciaError("AUTH_INVALID_REQUEST");
        }
        return sessionId;
    };
    createSessionCookies = (session) => {
        return this.sessionCookie.map((options) => createSessionCookie(session, this.env, options));
    };
}
